/*
** Module   : DisplayFormatFactory.java
** Abstract : factory for display formats.
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- --------------------------Description---------------------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SIY 20050901   @22535 Fixed formatting.
** 003 GES 20060711   @27975 Cleaned up code to use a common backing method.
** 004 SIY 20090701   @43016 Added format checking method which returns  detailed error
**                           information instead of throwing condition or displaying error
**                           immediately.
** 005 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 06 OM   20130329          Added datetime/datetime-tz and handle formats.
** 026 VMN 20140912          Fixes for format string validation and errors handling (ref.2382).
*/
package com.goldencode.p2j.ui.client.format;

import com.goldencode.p2j.ui.client.format.DisplayFormat.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;

/**
 * Factory for display formats. Because a display format is immutable it is
 * possible to make only one instance per format string value.
 */
public class DisplayFormatFactory
{
   /**
    * Dummy constructor to disable creation of instances of this class.
    */
   private DisplayFormatFactory() {}

   /**
    * Returns the default display format instance accordingly to type of the
    * given variable.
    * 
    * @param    var
    *           The variable to be represented.
    * 
    * @return   The default format for this variable type.
    * 
    * @throws   DisplayFormatParsingException
    *           A format instantiation error or the variable type is not
    *           supported.
    */
   public static DisplayFormat makeDefaultFormat(BaseDataType var)
   throws DisplayFormatParsingException
   {
      return makeFormat(var, var.defaultFormatString());
   }

   /**
    * Returns the display format instance according to the type of the given
    * variable and the format string.
    * 
    * @param    var
    *           The variable to be represented.
    * @param    pformat
    *           Valid format string.
    * 
    * @return   The format for this variable type configured with the given
    *           format string.
    * 
    * @throws   DisplayFormatParsingException
    *           A format instantiation error or the variable type is not
    *           supported.
    */
   public static DisplayFormat makeFormat(BaseDataType var, String pformat)
   throws DisplayFormatParsingException
   {
      DisplayFormat format;
      
      if (var instanceof character)
      {
         format = new StringFormat(pformat);
      }
      else if (var instanceof NumberType)
      {
         format = new NumberFormat(pformat, true);
      }
      else if (var instanceof datetimetz)
      {
         format = new DatetimeFormat(pformat, true);
      }
      else if (var instanceof datetime)
      {
         format = new DatetimeFormat(pformat, false);
      }
      else if (var instanceof date)
      {
         format = new DateFormat(pformat);
      }
      else if (var instanceof logical)
      {
         format = new BoolFormat(pformat);
      }
      else
      {
         String errmsg = "Unknown display format for variable's type: ";
         throw new DisplayFormatParsingException(errmsg + var.getClass().getName());
      }
      
      return format;
   }
   
   /**
    * Check validness of given value from the point of view of format.
    * 
    * @param    value
    *           Value to check.
    * @param    format
    *           Format to check.
    *           
    * @return   error code and message stored in common container.
    */
   public static ErrorEntry checkFormat(BaseDataType value, String format)
   {
      DisplayFormat fmt = null;
      try
      {
         if (format != null)
            fmt = DisplayFormatFactory.makeFormat(value, format);

      }
      catch (DisplayFormatParsingException e)
      {
         return new ErrorEntry(-1, "Unable to parse format \"" + format + "\"");
      }
      
      if (fmt != null)
      {
         Presentation pres = null;
         
         try
         {
            pres = fmt.fromVar(value);
         }
         catch (DisplayFormatCheckException e)
         {
            return new ErrorEntry(-1, "Unable to check format \"" + format + "\"" +
                                  " for value " + value.toStringMessage());
         }
         
         if (pres != null)
         {
            pres.finishEdit();
            return pres.validateFormat();
         }
      }
      
      return null;
   }
}
