/*
** Module   : NumberFormat.java
** Abstract : implements editing semantics for numbers
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 SVG 20050609   @21446 Initial version.
** 002 SVG 20050620   @21523 DisplayFormat related classes are redesigned.
**                           Now cursor's position and insert mode are
**                           maintained inside Presentation. This allows
**                           to not pass them by reference into
**                           Presentation edit methods. All
**                           com.goldencode.p2j.ui.chui.dataref package
**                           is obsoleted and removed.
** 003 SVG 20050803   @21993 Some fixes in export of integer variables.
** 004 SIY 20050831   @22537 Formatting and spelling fixes. 
** 005 SIY 20050907   @22642 Fixed use of @inheritDoc tag in javadoc
** 006 GES 20060115   @23917 Fixed loss of precision.
** 007 GES 20060119   @23978 Massive rewrite to leverage NumberType,
**                           decimal and integer for as much formatting,
**                           conversion and parsing as possible. This
**                           reduced the size and complexity of this
**                           class.  It also fixed a large number of
**                           bugs that were caused by the previous poor
**                           design.
** 008 ECF 20060214   @24724 Deactivated debug code. This is compiled out
**                           since it is now bracketed by the static DEBUG
**                           flag.
** 009 GES 20060505   @25983 Removed extra space on left of output. Only
**                           was visible to the user on left aligned
**                           fields, but it was present both types of 
**                           output.
** 010 GES 20060505   @25992 Added new feature to query the last valid
**                           cursor position.
** 011 GES 20060508   @26015 Ring the terminal bell on invalid input.
** 012 GES 20060511   @26073 Fix for all clearing a format with no 
**                           required digits.
** 013 GES 20060517   @26203 Exposed additional interfaces for proper
**                           numeric data formatting.  This is needed
**                           for cursor drawing and proper data display
**                           in zero suppressed formats.
** 014 GES 20060517   @26267 Fix for display of a zero in a zero- 
**                           suppressed rightmost integer position to the
**                           left of the decimal point. This is a case
**                           where the UI displays a digit when normal
**                           number formatting (e.g. the string built-in)
**                           would not.
** 015 GES 20060608   @27006 Support for '>' as a zero-suppressed digit.
** 016 GES 20060621   @27497 Fix for delete key (and some backspace key)
**                           failures. Cleaned up dead code and format.
** 017 GES 20060621   @27498 Fix for missing padding for right side sign
**                           when there is an explicit zero entered in a
**                           non-negative number.
** 018 GES 20060622   @27518 Fix to disallow user input into the "extra"
**                           digit position (the sign char position that
**                           is re-purposed for the display of a digit
**                           in some cases but which must not be accessed
**                           by the user while editing).
** 019 GES 20060711   @27976 New method for proper fixed numeric data
**                           formatting.
** 020 GES 20060724   @28154 Use forced assignment mode of BaseDataType
**                           to avoid checking the ErrorManager for a 
**                           pending error (that causes a trip to the
**                           server which is unnecessary).
** 021 GES 20060811   @28552 Created a more correct algorithm for
**                           isFixedNumeric() that seems to match the
**                           4GL behavior exactly.
** 022 GES 20060829   @28951 Added proper handling of the decimal and
**                           group separators in input().  Previously
**                           there was no support at all (it just rang
**                           the bell).
** 023 GES 20060830   @28988 Switch use of decimal.toString to a version
**                           that disables automatic left padding.
** 024 GES 20060920   @29730 Added support for isUnknown().
** 025 GES 20060927   @29965 Fixed overflow detection in cases where there
**                           is no sign in the format (such cases were
**                           being treated as if they had an optional left
**                           sign).
** 026 NVS 20061026   @30701 Added support for separate left side and
**                           right side signs since both can be used at
**                           the same time.
** 027 SIY 20070425   @33227 Fixed handling of formats like -zz,zzz.99.
** 028 SIY 20070720   @34614 Fixed handling of right-aligned formats for
**                           unknown value.
**                           Ref: #34615
** 029 SIY 20070808   @34835 Fixed reporting of alignment.
** 030 SIY 20071015   @35473 Rolled back part of #028. It is not necessary
**                           now and causes regressions for some 
**                           combinations of format and label placement.
**                           Ref: #35474
** 031 SIY 20071204   @36211 Fixed alignment of formats like "abc->>>>9".
**                           Credits to CA for nailing it down.
**                           Ref: #36210
** 032 CA  20080114   @36809 Added method isSplitFormat.
** 033 CA  20080115   @36811 Fixed displaying numeric values with format  
**                           (>>>): the leading space for positive values 
**                           must not be eliminated.
** 034 EVL 20080415   @37992 Adding handling for the special case when
**                           the value is 0 but the sign is negative. We
**                           must have the screen value -0.0 for this
**                           case.
** 035 GES 20080416   @38011 Changed methods by which we initialize the
**                           decimal and group separators to allow the
**                           result to be controlled by the directory.
** 036 NVS 20080611   @38688 Avoid double error messages for the same
**                           format error.
** 037 NVS 20080704   @39066 Fixed BACKSPACE function for left-aligned
**                           fields, which didn't reposition the cursor
**                           properly.
** 038 CA  20080819   @39434 Fixed BACKSPACE for left-aligned fields with
**                           a format like '>>>>>>9' (at least one digit
**                           is specified).
** 039 SVL 20080914   @39797 Insert mode reset removed from the
**                           activate().
** 040 CA  20090122   @41196 "z" format quirk - allow leading digits if,
**                           after deleting the first digit in a number,
**                           the next digits are 0: i.e. "10000" will
**                           display "0000" after 1 is deleted.
** 041 GES 20090424   @41965 Import change.
** 042 SVL 20090612   @42679 Call init() into NumberBuf(NumberType) even if we
**                           have an overflow.
** 043 SIY 20090624   @42888 Fixed positioning when "?" is typed as input.
** 044 SIY 20090721   @43293 Refactored initial cursor position calculation
**                           into separate method.
** 045 GES 20090813   @43612 Added safety code to avoid index out of bounds
**                           exceptions during rendering. The source data and
**                           format string are unknown at this time. That data
**                           is logged to STDERR to allow notification of that
**                           case when it happens again.
** 046 SIY 20090909   @43853 Fixed OOB exception and removed warning in OOB
**                           protection added in #45.
** 047 EVL 20100406   @44793 Special processing for ? is required to fix the
**                           insert mode issue.
** 048 SIY 20100619          Switching to own implementation of character UI.
** 049 SIY 20100804          Changes in imports, some refactorings.
** 050 EVL 20100909          The initArray() method has been changed to include
**                           the sign from the right side consideration in
**                           extra digit artificial adding.
** 051 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 052 OM  20130406          Added int64 variable support, added temporary client-side
**                           validation for integer/int64.
** 005 CA  20131001          Eliminated custom group/dec separators when format is used. In this
**                           case, they must default to comma (,) for group and dot (.) for
**                           decimal separator. The custom group/dec separators must be used only
**                           for toString/parsing purposes.
** 006 VMN 20140912          Fixes for format string validation and errors handling (ref.2382).
*/

package com.goldencode.p2j.ui.client.format;

import com.goldencode.util.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Display format processing for <code>integer</code>, <code>int64</code> and <code>decimal</code>
 * variables.
 */
public class NumberFormat
extends DisplayFormat
{
   /** Allows quick comparisons with zero. */
   private static final decimal DECIMALZERO = new decimal(0);
   
   /** Stores parsed format string metrics. */
   protected NumberType.FormatSpec fs = null; 

   /** Is decimal period present? */
   protected boolean fmtHavePeriod = false;

   /** Number of '&gt;' chars. */
   protected int fmtIntSharedCount = 0;

   /** Number of 'Z' or '9' or '*' chars before period. */
   protected int fmtIntDigitsCount = 0;

   /** Number of '9' chars before period. */
   protected int fmtIntZerosCount = 0;

   /** Number of '9' chars after period. */
   protected int fmtFractDigitsCount = 0;

   /** Number of '&lt;' chars. */
   protected int fmtFractSharedCount = 0;

   /** Default width of widget's content. */
   protected int fmtScreenWidth = 0;

   /** Control format alignment. */
   private boolean rightAligned = true;

   /** Complex format with split alignment. */
   private boolean splitFormat = false;

   /** Clear leading zeros for Z format. */
   private boolean clearLeadingZeros = false;
   
   /**
    * Constructs a number (decimal, int64 or integer) format from a Progress
    * compatible format string.
    * 
    * @param   pformat
    *          Progress format string.
    * @param   isThrow
    *          The way of generating an error. If true, ErrorManager.recordOrThrowError() will
    *          be invoked, otherwise ErrorManager.recordOrShowError().
    *
    * @throws  DisplayFormatParsingException
    *          Format string is invalid.
    */
   public NumberFormat(String pformat, boolean isThrow)
   throws DisplayFormatParsingException
   {
      super(pformat);

      fs = NumberType.parseFormat(pformat, isThrow);
      
      fmtHavePeriod = (fs.decimalPoint != -1);
      
      int len = (fs.digitsLeft != null) ? fs.digitsLeft.length() : 0;
      
      // calculate metrics for the left side
      for (int i = 0; i < len; i++)
      {
         char ch = fs.digitsLeft.charAt(i);
         
         switch (ch)
         {
            case '9':
               fmtIntZerosCount++;
               // intentionally falls through!
            case 'Z':
            case 'z':
            case '*':
               fmtIntDigitsCount++;
               break;
            case '>':
               fmtIntSharedCount++;
               rightAligned = false;
               break;
         }
      }
      
      // NOTE: At present we set this flag for limited number of cases, 
      // in order to minimize possible regressions, but switching it to true 
      // permanently might be also correct. 
      if (fs.userLeft.length() > 0 && !rightAligned)
         splitFormat = true;
      
      // calculate metrics for the right side
      len = (fs.digitsRight != null) ? fs.digitsRight.length() : 0;
      
      for (int i = 0; i < len; i++)
      {
         char ch = fs.digitsRight.charAt(i);
         
         switch (ch)
         {
            case '9':
               fmtFractDigitsCount++;
               break;
            case '<':
               fmtFractSharedCount++;
               break;
         }
      }
            
      // calculate the width of the sign(s)
      int signWidth = signSize(fs.signLeft) + signSize(fs.signRight);
      
      // calculate the screen width
      fmtScreenWidth = fs.userLeft.length()    +
                       signWidth               +
                       fs.digitsLeft.length()  +
                       fs.userRight.length();
                       
      if (fmtHavePeriod)
      {
         // including decimal period and digits after it
         fmtScreenWidth += fs.digitsRight.length() + 1;
      }
   }

   /**
    * Returns screen width of the format.  The screen width is the number of
    * characters that will be rendered.  This is used as the default size for
    * an editor.
    * 
    * @return   The screen width of the format. 
    */
   @Override
   public int getScreenWidth()
   {
      return fmtScreenWidth;
   }

   /**
    * Indicates if the screen value or a column's title could be formatted
    * with trimming from the left and right alignment based on this format.
    * <p>
    * This value only notes if this format supports such trimming/alignment
    * but it doesn't specify if it should be applied in every case (which
    * may be a configuration option for a given caller).
    * 
    * @return   Always <code>true</code>. 
    */
   @Override
   public boolean isRightPositioning()
   {
      return rightAligned ;
   }

   /**
    * Indicates if current output is split format
    * 
    * @return   value of the splitFormat field. 
    */
   public boolean isSplitFormat()
   {
      return splitFormat;
   }

   /**
    * Returns a new presentation instance with a default initialized insert
    * mode and cursor position.
    * 
    * @param   pres
    *          A Presentation instance which was constructed by this format.
    *          May be <code>null</code>. This presentation instance is not 
    *          changed in this method.
    * 
    * @return  New presentation instance.
    */
   @Override
   public Presentation clear(Presentation pres)
   {
      NumberBuf nb = new NumberBuf();
      
      if (pres != null && pres instanceof NumberBuf)
      {
         NumberBuf old = (NumberBuf) pres;
         nb.justActivated = old.justActivated;
      }
         
      return nb;
   }

   /**
    * Initializes a presentation for variable's value.
    * 
    * @param   var
    *          Variable which should be compatible with format.
    * 
    * @return  Data presentation
    * 
    * @throws  DisplayFormatCheckException
    *          Variable's value is incompatible with format.
    */
   @Override
   public Presentation fromVar(BaseDataType var)
   throws DisplayFormatCheckException
   {
      return new NumberBuf((NumberType)var);
   }

   /**
    * Calculates the number of characters in a sign representation.
    * 
    * @param   signType
    *          encoding of sign type
    * 
    * @return  numbers of characters to represent sign of this type
    */
   private int signSize(int signType)
   {
      if (signType == NumberType.SIGN_NONE)
         return 0;
      
      return (signType == NumberType.SIGN_PAREN       || 
              signType == NumberType.SIGN_PAREN_EMBED ||
              signType == NumberType.SIGN_CR          ||
              signType == NumberType.SIGN_DR          ||
              signType == NumberType.SIGN_DB) ? 2 : 1;
   }

   /**
    * Internal form of a number.
    */
   private class NumberBuf
   extends Presentation
   {
      /** Special "digit" for explicit leading zero. */
      private static final byte LEADING_ZERO = -1;
      
      /**
       * Array of digits. Last <code>scale</code> of them are after decimal
       * point. Several first bytes may have value <code>LEADING_ZERO</code>.
       * This special value is necessary to show something like " 00023" or
       * "0" with "z(8)" format. May be <code>null</code> which means the
       * <code>unknown value</code>.
       */
      private byte[] digits = null;

      /**
       * Current decimal scale for <code>digits</code> array. May be changed
       * during edit if format includes '&lt;' elements.
       */
      private int scale;

      /** Does number contain negative sign. */ 
      private boolean negative;

      /** Has an overflow (number too big to fit in format) occurred? */ 
      private boolean overflow;

      /** If extra char is possible due to certain format sign options. */ 
      private boolean extra;
      
      /** Flags if activation just occurred (flag resets on any input). */
      private boolean justActivated = false;

      /**
       * Cached screen rendering info. It always contains fmtScreenWidth
       * elements. Number of used elements is screen.length(). To be thread
       * safe this member should be used for synchronization.
       */
      private ScreenCharInfo[] screenInfo;

      /** Cached screen value. */
      private String screen;

      /** Buffer for screen value. Used only during calculations. */
      private StringBuffer screenBuf;

      /**
       * Constructs default presentation used for clear operation.
       */
      public NumberBuf()
      {
         initClear();
         presInsertMode = true;
      }

      /**
       * Constructs presentation initialized with given <code>decimal</code>, <code>int64</code> 
       * or <code>integer</code>.
       * 
       * @param   var
       *          Wrapper containing our initial value and which will be
       *          used to format and parse the data.
       */
      public NumberBuf(NumberType var)
      {
         // leave the byte array null if the value is unknown
         if (var.isUnknown())
         {
            return;
         }
         
         // initialize the array
         initArray();
         
         // format string
         String fmt = NumberFormat.this.getFormatDef();
         
         boolean unsigned = (fs.signLeft  == NumberType.SIGN_NONE &&
                             fs.signRight == NumberType.SIGN_NONE);
         
         // detect if the value is negative
         if (CompareOps.lessThan(var, DECIMALZERO))
         {
            if (unsigned)
            {
               // this will trigger error message display/logging
               // and return here
               var.toString(fmt);
               init();
               overflow = true;
               return;
            }

            negative = true;
         }
         else
         {
            negative = false;
         }

         // cut the format down to a version that can be easily parsed (no
         // user text or group separators, sign chars are normalized too)
         String simpleFmt = NumberType.generateSimpleFormat(fmt, !unsigned);
         
         // obtain the formatted data
         String text = var.toString(simpleFmt);
                  
         // number is too big to fit in format, record the overflow
         if (text.indexOf('?') > -1)
         {
            init();

            overflow = true;
            return;
         }
         
         // remove any sign or leading space
         if (negative || text.charAt(0) == ' ')
         {
            text = text.substring(1);
         }         
         
         // split the result into left and right (of the decimal point) sides
         char decSep = NumberType.getDecimalSeparator();
         int    point = text.indexOf(decSep);
         String left  = (point >  0) ? text.substring(0, point)
                                     : (point == 0 ? "" : text);
         String right = (point >= 0) ? text.substring(point + 1) : "";
         
         // calculate the scale
         scale = right.length();

         // the index into our byte array (we start writing from the end and
         // work backward)
         int idx = digits.length - 1;
         
         // copy the right side chars, from back to front
         for (int i = right.length() - 1; i >= 0; i--)
         {
            digits[idx] = (byte) Character.digit(right.charAt(i), 10);
            idx--;
         }

         // copy the left side chars, from back to front
         for (int i = left.length() - 1; i >= 0; i--)
         {
            digits[idx] = (byte) Character.digit(left.charAt(i), 10);
            idx--;
         }

         init();
      }

      /**
       * Returns the last valid cursor position in the current data.  This
       * is equivalent to the number of actual character positions in the 
       * data.
       * 
       * @return  The last valid cursor position.
       */
      @Override
      public int getLastPos()
      {
         validateScreen();
         return (screen.length() - 1);
      }
      
      /**
       * Returns the sign state of the numeric value.
       *
       * @return  <code>true</code> if the value is negative.
       */
      @Override
      public boolean isNegative()
      {
         return negative;
      }
   
      /**
       * Detects if the numeric sign formatting character exists on the left
       * in the format and if it is optional.
       *
       * @return   <code>true</code> if the format has an optional sign 
       *           character.
       */
      @Override
      public boolean isOptionalLeftSign()
      {
         return (fs.signLeft == NumberType.SIGN_LEFT_MINUS ||
                 fs.signLeft == NumberType.SIGN_PAREN);
      }

      /**
       * Detects if the numeric format is comprised of digits that always
       * output (like "-99,999").
       *
       * @return   <code>true</code> if all digits in on the left or right
       *           of the decimal point always generate a corresponding
       *           character in the output. This means that no "&gt;" or
       *           "&lt;" characters exist in the format.
       */
      @Override
      public boolean isFixedNumeric()
      {
         return ((fmtIntSharedCount + fmtFractSharedCount) == 0);
      }
      
      /**
       * Returns the size of the right side sign formatting characters in the
       * format.
       *
       * @return   The number of right side sign characters.
       */
      public int rightSignSize()
      {
         int num = 0;
         
         switch (fs.signRight)
         {
            case NumberType.SIGN_NONE:
               if (fs.signLeft == NumberType.SIGN_PAREN ||
                   fs.signLeft == NumberType.SIGN_PAREN_EMBED)
                  num = 1;
               break;
            case NumberType.SIGN_RIGHT_MINUS:
            case NumberType.SIGN_RIGHT_PLUS:
               num = 1;
               break;
            case NumberType.SIGN_CR:
            case NumberType.SIGN_DB:
            case NumberType.SIGN_DR:                 
               num = 2;
               break;
         }
         
         return num;
      }

      /**
       * Performs the necessary initialization for a widget which is
       * activated by a focus change (using a tab).
       */
      @Override
      public void activate()
      {
         justActivated = true;
         super.activate();
      }
      
      /**
       * Get initial cursor position which will be set during activation.
       * 
       * @return  calculated cursor position.
       */
      @Override
      public int getInitialCursorPos()
      {
         int res = 0;
         
         if (digits != null)
         {
            validateScreen();
            
            boolean leftDigits = false;

            // set the cursor position to the leftmost digit char
            for (int cp = 0; cp < screenInfo.length; cp++)
            {
               int type = screenInfo[cp].getType();
               
               if (cp >= screen.length())
                  break;
               
               if (type == ScreenCharInfo.DIGIT)
               { 
                  leftDigits = true;
                  
                  // saving digit position
                  res = cp;
                  
                  if (Character.isDigit(screen.charAt(cp)))
                  { 
                     // suitable digit found
                     break;
                  }
               }
               else if (type == ScreenCharInfo.DEC_SEP)
               {
                  // the default position is on left of decimal point but if
                  // there are no digits to the left, then the first char
                  // after the decimal point is the cursor position
                  if (!leftDigits)
                  {
                     res = cp + 1;
                  }
                  
                  break;
               }
            }
         }
         else
            if (isRightPositioning())
               res = getScreenWidth() - 1;
         
         return res;
      }

      /**
       * Checks format rules. Method finishEdit() should be called before it.
       * <p>
       * This is a no-operation method in this class. It is not needed since
       * a number presentation is always maintained in the correct state.
       */
      @Override
      public void checkFormat()
      {
      }

      /**
       * Returns if the presentation currently represents the 
       * <code>unknown value</code>.  If any editing has occurred, then this
       * will return <code>false</code> even if the editing is not complete.
       *
       * @return   <code>true</code> if this is <code>unknown</code>.
       */
      @Override
      public boolean isUnknown()
      {
         return (digits == null);
      }
      
      /**
       * Performs BACKSPACE function which consists of deleting one character
       * at the cursor and then moving the cursor one position to the left,
       * if possible.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteBack()
      {
         if (digits == null)
         {
            return false;
         }

         boolean res = true;

         if (presInsertMode)
         {
            if (presCursorPos > 0 || screen.length() == 1)
            {
               res = deleteCurrent();

               // deleteCurrent will reposition the cursor for formats like
               // '>>>>>>9' (when fmtIntDigitsCount > 0) - so, no 
               // decrementing is necessary for this format
               if (fmtIntDigitsCount == 0 && 
                   !rightAligned          && 
                   presCursorPos > 0)
               {
                  presCursorPos --;
               }
            }
         }
         else
         {
            res = moveCursorLeft();
         }

         return res;
      }

      /**
       * Erases one symbol in the cursor position.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean deleteCurrent()
      {
         if (digits == null)
         {
            return false;
         }

         validateScreen();

         if (presCursorPos >= screen.length() ||
             screenInfo[presCursorPos].getType() != ScreenCharInfo.DIGIT)
         {
            return false;
         }

         int digitPos = screenInfo[presCursorPos].getDigitPos();
         int digitIndex = calcDigitIndex(digitPos);

         if (digits[digitIndex] == LEADING_ZERO)
         {
            return false;
         }

         // shift digits right 1 digit position
         for (int i = digitIndex; i > 0; i--)
         {
            digits[i] = digits[i - 1];
         }
         digits[0] = LEADING_ZERO;

         if (digitPos < 0)
         { 
            // digit removed in fractional part
            int firstFractIndex = digits.length - scale;
            
            if (digits[firstFractIndex] == LEADING_ZERO)
            { 
               // leading zero may be only in integer-type part
               digits[firstFractIndex] = 0;
            }
         }
         else if (digits[digitIndex] == LEADING_ZERO)
         { 
            // first digit erased, clearing next zeros to free left dynamic 
            // area
            int firstFixedIndex = calcDigitIndex(fmtIntDigitsCount - 1);

            for (int i = digitIndex + 1; i < firstFixedIndex; ++i)
            {
               if (digits[i] != 0)
               {
                  break;
               }
               digits[i] = LEADING_ZERO;
            }
         }

         int zeroesBeginIndex = digits.length - scale - fmtIntZerosCount;
         
         // scale and fmtIntZerosCount can both be zero, if so then we don't
         // need to check here (and we can't since it would generate an
         // array index out of bounds)
         if (zeroesBeginIndex < digits.length &&
             digits[zeroesBeginIndex] == LEADING_ZERO)
         {
            // "99..9" part must not contain leading zeros
            digits[zeroesBeginIndex] = 0;
         }

         renderScreen();

         // digit position may disappear or be shifted so we must search for
         // a new cursor position, with fmtIntDigitsCount > 0 we can guarantee
         // that we have the same position or at least one after it on the
         // same side of decimal point (possibly filled with LEADING_ZERO)
         if (fmtIntDigitsCount > 0)
         {
            int newCursorPos = 0;
            int newDigitPos = -1;
            
            while (newCursorPos < screen.length())
            {
               ScreenCharInfo info = screenInfo[newCursorPos];
               if (info.getType() == ScreenCharInfo.DIGIT &&
                   info.getDigitPos() <= digitPos)
               {
                  newDigitPos = info.getDigitPos();
                  break;
               }
   
               newCursorPos++;
            }
   
            if (newDigitPos != -1)
            {
               int newDigitIndex = calcDigitIndex(newDigitPos);
               
               if (digits[newDigitIndex] == LEADING_ZERO)
               {
                  // we are in integer-type part and need to skip some new leading
                  // zeros but still stay left of the decimal point
                  while (newDigitPos > 0)
                  {
                     newCursorPos++;
                     ScreenCharInfo info = screenInfo[newCursorPos];
                     
                     if (info.getType() == ScreenCharInfo.DIGIT)
                     {
                        newDigitPos--;
                        newDigitIndex++;
                        if (digits[newDigitIndex] != LEADING_ZERO)
                        {
                           break;
                        }
                     }
                  }
               }
            }
   
            presCursorPos = newCursorPos;
         }
         
         return true;
      }

      /**
       * Inputs character. The result depends from the current insert mode.
       * 
       * @param   ch
       *          Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean input(char ch)
      {
         char decSep = NumberType.getDecimalSeparator();
         char groupSep = NumberType.getGroupSeparator();

         clearLeadingZeros = false;

         if (ch == '?')
         {
            // Synchronization the data presentation insert mode
            setInsertMode(Window.getInsertMode() == Window.INSERT_MODE_INS);
            // Note that this differs from the initial '?' formatting. 
            // In this case '?' is always left aligned and cursor positioned
            // to '?'.
            digits = null;
            invalidateScreen();
            
            presCursorPos = 0;
            screen = StringHelper.leftAlignText("?", getScreenWidth());
            
            return true;
         }

         if (digits == null)
         { 
            // restarting method from zero content
            initClear();
            boolean res = input(ch);
            
            // if input was not accepted then we reset the state
            if (!res)
            {
               digits = null;
               invalidateScreen(); // cache should cleared
            }
            return res;
         }

         validateScreen();

         if (ch == groupSep)
         {
            // formats with no decimal point always ring bell even if the
            // format has a group separator
            if (!fmtHavePeriod)
            {
               OutputManager.instance().beep();
               return false;
            }
            
            // the group separator is always silently ignored for decimals
            // (no beep, no change in cursor position or contents), although
            // if this is the first input after activation, the cursor will
            // move to the least significant integer position
            if (justActivated)
            {
               // the insert mode is set already by the fill-in event
               // processing
               
               // the cursor movement may have been set by activation 
               // processing and/or a call to clear(), if so we just allow
               // it to be honored BUT if our cursor position is 0 then
               // we try to move it
               if (presCursorPos == 0)
               {
                  for (int i = 0; i < screen.length(); i++)
                  {
                     int type = screenInfo[i].getType();
                     
                     if (type == ScreenCharInfo.DEC_SEP)
                     {
                        // move cursor to the least significant integral digit
                        // as long as there is such a digit 
                        if (i > 0)
                           presCursorPos = i - 1;
                        
                        return true;
                     }
                  }
               }
               
               justActivated = false;
               return true;
            }
            return false;
         }
         else if (ch == decSep)
         {
            // the decimal separator can be input in the following cases
            // (formats with a decimal point only):
            // - the present cursor position is on the decimal separator
            //   position (in the current format string), regardless of
            //   whether the field is in insert or overwrite mode
            // - in insert mode, if the present cursor position is one  
            //   position left of the decimal separator position (in the
            //   current format string)
            // - the field was just activated and the cursor is in the 
            //   leftmost position and this is overwrite mode 
            if (presCursorPos >= 0               &&
                presCursorPos < screen.length()  &&
                fmtHavePeriod)
            {
               int num  = 0;
               int type = screenInfo[presCursorPos].getType();
               
               if (type == ScreenCharInfo.DEC_SEP)
               {
                  num = 1;
               }
               else if (presInsertMode)
               {
                  if (presCursorPos < screen.length() - 1)
                  {
                     type = screenInfo[presCursorPos + 1].getType();
                     
                     if (type == ScreenCharInfo.DEC_SEP)
                     {
                        num = 2;
                     }
                  }
               }
               else if (justActivated && presCursorPos == 0)
               {
                  // assumes that this is not insert mode due to order of the
                  // "else if" clauses above
                  for (int i = 0; i < screen.length(); i++)
                  {
                     type = screenInfo[i].getType();
                     
                     if (type == ScreenCharInfo.DEC_SEP)
                     {
                        num = i + 1;
                        break;
                     }
                  }
               }
      
               // check if one of the valid cases occurred
               if (num > 0)
               {
                  justActivated = false;
                  
                  // reset input mode to overwrite
                  presInsertMode = false;
                  
                  // move the cursor to the fractional part
                  presCursorPos += num;
                  
                  return true;
               }
            }               
            
            OutputManager.instance().beep();
            return false;
         }
         else if (ch == ' ')
         {
            if (presCursorPos >= screen.length() ||
                screenInfo[presCursorPos].getType() !=
                ScreenCharInfo.DIGIT)
            {
               return false;
            }

            int digitPos = screenInfo[presCursorPos].getDigitPos();
            int digitIndex = calcDigitIndex(digitPos);
            
            if (digitPos < 0 || digits[digitIndex] == LEADING_ZERO ||
                digitIndex > 0 && digits[digitIndex - 1] != LEADING_ZERO)
            {
               return false;
            }

            justActivated = false;

            // we are over a first digit so we can replace it by leading
            // zero
            digits[digitIndex] = LEADING_ZERO;

            if (digitPos < fmtIntDigitsCount)
            {
               // only current char changed so cursor will not move
               invalidateScreen();
               return true;
            }
            // clean next leading zeros
            int firstFixedIndex = calcDigitIndex(fmtIntDigitsCount - 1);
            for (int i = digitIndex + 1; i < firstFixedIndex; i++)
            {
               if (digits[i] != 0)
               {
                  break;
               }
               digits[i] = LEADING_ZERO;
            }
            
            renderScreen();
            
            // searching for new cursor position
            for (int newCursorPos = 0; newCursorPos < screen.length(); 
                 newCursorPos++)
            {
               ScreenCharInfo info = screenInfo[newCursorPos];
               if (info.getType() == ScreenCharInfo.DIGIT)
               {
                  int newDigitPos = info.getDigitPos();
                  if (newDigitPos < digitPos || newDigitPos == 0)
                  {
                     presCursorPos = newCursorPos;
                     break;
                  }
               }
            }
            
            return true;
         }
         else if (ch == '+')
         {
            clearLeadingZeros = true;
            if (negative)
            {
               // optional sign is now gone (the digit position is being
               // removed), shift cursor to ensure it stays on the current
               // digit
               if (isOptionalLeftSign())
                  presCursorPos--;
               
               negative = false;
               invalidateScreen();
            }
            justActivated = false;
            return true;
         }
         else if (ch == '-')
         {
            clearLeadingZeros = true;
            if (fs.signLeft  == NumberType.SIGN_NONE &&
                fs.signRight == NumberType.SIGN_NONE)
            {
               OutputManager.instance().beep();
               return false;
            }

            if (!negative)
            {
               // optional sign is now present (a new digit position is being
               // added), shift cursor to ensure it stays on the current
               // digit
               if (isOptionalLeftSign())
                  presCursorPos++;
               
               negative = true;
               invalidateScreen();
            }
            justActivated = false;
            return true;
         }

         boolean result = (presInsertMode) ? insert(ch) : overwrite(ch);
         
         if (!result)
         {
            // on invalid input, ring the terminal bell
            OutputManager.instance().beep();
         }
         else
         {
            // valid input
            justActivated = false;
         }
         
         return result;
      }

      /**
       * Moves cursor one character to the left.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorLeft()
      {
         if (presCursorPos > 0)
         {
            presCursorPos--;
         }
         return true;
      }

      /**
       * Moves cursor one character to the right. Also updates insert mode.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorRight()
      {
         if (digits == null)
         {
            if (presCursorPos < getScreenWidth() - 1)
            {
               presCursorPos++;
            }
            return true;
         }

         validateScreen();

         if (presCursorPos >= screen.length() - 1)
         {
            return false;
         }

         if (screenInfo[presCursorPos].getType() == ScreenCharInfo.DEC_SEP)
         {
            // moving to fractional part resets input mode to overwrite
            presInsertMode = false;
         }

         presCursorPos++;
         return true;
      }

      /**
       * Moves cursor to the beginning of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorBegin()
      {
         presCursorPos = 0;
         return true;
      }

      /**
       * Moves cursor to the end of the field.
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      @Override
      public boolean moveCursorEnd()
      {
         presCursorPos = toScreenValue().length() - 1;
         return true;
      }

      /**
       * Indicates completion of an edit. The presentation instance may fill 
       * some unfilled places.
       * 
       * @return   <code>true</code> if rendering is necessary.
       */
      @Override
      public boolean finishEdit()
      {
         return false; // dummy method
      }

      /**
       * Formats the data as a string that is ready for display.
       * 
       * @return   The formatted string.
       */
      @Override
      public String toScreenValue()
      {
         validateScreen();
         return screen;
      }

      /**
       * Instantiates a variable that is equivalent to the presentation's
       * internal representation.
       * 
       * @param    var
       *           A variable which must be compatible with the format.
       */
      @Override
      public void toVar(BaseDataType var)
      {
         if (var instanceof integer)
         {
            // first check the derived class before base
            toIntegerVar((integer) var);
         }
         else if (var instanceof int64)
         {
            toInt64Var((int64) var);
         }
         else
         {
            toDecimalVar((decimal) var);
         }
      }

      /**
       * Common part in constructors.
       */
      private void init()
      {
         int dgtEndIndex = digits.length - scale - fmtIntZerosCount;
         
         for (int i = 0; i < dgtEndIndex; i++)
         {
            if (digits[i] != 0)
            {
               break;
            }
            digits[i] = LEADING_ZERO;
         }

         screenInfo = new ScreenCharInfo[fmtScreenWidth];
         
         for (int i = 0; i < fmtScreenWidth; i++)
         {
            screenInfo[i] = new ScreenCharInfo();
         }

         screenBuf = new StringBuffer(fmtScreenWidth);
      }

      /**
       * Resets content to zero value and sets cursor to the first zero digit.
       */
      private void initClear()
      {
         // reserving for fixed fractional digits
         scale = fmtFractDigitsCount;
         
         negative = false;
         
         // initialize the array
         initArray();
         
         // final initialization
         init();

         renderScreen();
         
         for (int pos = 0; pos < screen.length(); pos++)
         {
            ScreenCharInfo info = screenInfo[pos];

            // include case with something like .99 after 
            // zero-suppressed format strings 
            if ((info.getType() == ScreenCharInfo.DIGIT || 
                 (info.getType() == ScreenCharInfo.DEC_SEP && 
                  fmtHavePeriod && isZeroSuppressed())) &&
                 info.getDigitPos() == 0)
            {
               presCursorPos = pos;
               return;
            }
         }

         // format strings like "zzzzz", "-zzz,zzz,zzz" or "*****" will have
         // no actual digits in the result, we need to set the position to 0
         // in such cases (only can happen in formats without a decimal point 
         // since z and * can't occur to the right of the decimal)
         
         if (!fmtHavePeriod && isZeroSuppressed())
         {
            presCursorPos = 0;
         }
         else
         {
            throw new IllegalStateException("Can't init zero number for " +
                                            getFormatDef() + " format");
         }
      }
      
      /**
       * Determines if the format string has all integer digits (digits to
       * the right of the decimal point) that are zero-suppressed (as a
       * 'z' or 'Z' or '*' or '>').  If any left digit is a '9', the format 
       * is not zero suppressed.
       *
       * @return   <code>true</code> if all left digits are zero-suppressed.
       */
      private boolean isZeroSuppressed()
      {
         int left          = fs.digitsLeft.length();
         int leftDigits    = (fs.seps > 0) ? (left - fs.seps) : left; 
         int nonZeroDigits = fmtIntDigitsCount + fmtIntSharedCount;
         
         return (fmtIntZerosCount == 0 && nonZeroDigits == leftDigits);
      }
      
      /**
       * Determines if there are no digits on the right of any decimal point.
       *
       * @return   <code>true</code> if there are no digits on the right of
       *           the decimal point.
       */
      private boolean isLeftOnly()
      {
         return (fs.digitsRight == null || fs.digitsRight.length() == 0);
      }

      /**
       * Calculates an index in the digits array corresponding to the given
       * digit decimal position.
       * 
       * @param   digitPos
       *          Digit position i.e. decimal shift
       * 
       * @return  Index in digits array for it.
       */
      private int calcDigitIndex(int digitPos)
      {
         return (digits.length - 1) - scale - digitPos;
      }

      /**
       * Places current value into a <code>integer</code> variable.
       * 
       * @param    var
       *           Target variable.
       */
      private void toIntegerVar(integer var)
      {
         if (digits == null)
         {
            var.setUnknown();
            return;
         }

         int res = 0;
         int firstFractIndex = digits.length - scale;
         
         // walk through the data and convert each digit
         for (int i = 0; i < firstFractIndex; i++)
         {
            // bypass leading zeroes
            if (digits[i] == LEADING_ZERO)
            {
               continue;
            }
            
            if ((res > Integer.MAX_VALUE / 10) ||
                (res == Integer.MAX_VALUE / 10 && digits[i] > Integer.MAX_VALUE % 10))
            {
               // at the current iteration the res value will overflow, we stop here
               ErrorManager.recordOrThrowError(
                       15747,
                       "Value " + toScreenValue() + " too large to fit in INTEGER");
               // TODO: OM: After a discussion with Constantin, I understand that the validation
               // is not well placed here (this executes on client-side), however, at this moment
               // the container for transporting the buffer to server is a variable of the exact 
               // type, so we must parse/validate before putting the value into container.
               // When this will be changed to actual char (?) buffer, these validations will 
               // become useless and should be removed along with next TODO comments.
               break;
            }
            res = res * 10 + digits[i];
         }

         // round up if needed
         if (scale > 0 && digits[firstFractIndex] >= 5)
         {
            // TODO: if res is already Integer.MAX_VALUE, this line will overflow
            res++;
         }

         // set the sign
         if (negative)
         {
            // TODO: this algorithm does not allow to read Integer.MIN_VALUE because the 
            // -Integer.MIN_VALUE is larger than Integer.MAX_VALUE and cannot fit in the 
            // available bits
            res = -res;
         }

         var.setValue(res);
      }

      /**
       * Places current value into a <code>int64</code> variable.
       * 
       * @param    var
       *           Target variable.
       */
      private void toInt64Var(int64 var)
      {
         if (digits == null)
         {
            var.setUnknown();
            return;
         }

         long res = 0;
         int firstFractIndex = digits.length - scale;
         
         // walk through the data and convert each digit
         for (int i = 0; i < firstFractIndex; i++)
         {
            // bypass leading zeroes
            if (digits[i] == LEADING_ZERO)
            {
               continue;
            }
            
            if ((res > Long.MAX_VALUE / 10) ||
                (res == Long.MAX_VALUE / 10 && digits[i] > Long.MAX_VALUE % 10))
            {
               // at the current iteration the res value will overflow, we stop here
               ErrorManager.recordOrThrowError(78, "Value too large for integer.");
               // TODO: OM: After a discussion with Constantin, I understand that the validation
               // is not well placed here (this executes on client-side), however, at this moment
               // the container for transporting the buffer to server is a variable of the exact 
               // type, so we must parse/validate before putting the value into container.
               // When this will be changed to actual char (?) buffer, these validations will 
               // become useless and should be removed along with next TODO comments.
               break; 
            }

            res = res * 10 + digits[i];
         }

         // round up if needed
         if (scale > 0 && digits[firstFractIndex] >= 5)
         {
            // TODO: if res is already Integer.MAX_VALUE, this line will overflow
            res++;
         }

         // set the sign
         if (negative)
         {
            // TODO: this algorithm does not allow to read Long.MIN_VALUE because the
            // -Long.MIN_VALUE is larger than Long.MAX_VALUE and cannot fit in the available
            // bits
            res = -res;
         }

         var.setValue(res);
      }

      /**
       * Places current value into a <code>decimal</code> variable.
       * 
       * @param    var
       *           Target variable.
       */
      private void toDecimalVar(decimal var)
      {
         if (digits == null)
         {
            var.setUnknown();
            return;
         }

         char decSep = NumberType.getDecimalSeparator();

         StringBuffer buf = new StringBuffer(digits.length + 1);
         
         if (negative)
         {
            buf.append('-');
         }
         
         int point = digits.length - 1 - scale;
         
         for (int i = 0; i < digits.length; i++)
         {
            if (digits[i] != LEADING_ZERO)
            {
               buf.append(Character.forDigit(digits[i], 10));
            }
            else
            {
               buf.append('0');
            }
            
            if (i == point)
            {
               buf.append(decSep);
            }
         }

         var.assign(new decimal(buf.toString()), true);
      }

      /**
       * Inputs a character in insert mode.
       * 
       * @param    ch
       *           Input character
       * 
       * @return   <code>true</code> if input is accepted (note that it
       *           may be silently ignored in reality).
       */
      private boolean insert(char ch)
      {
         // accept only digits
         if (!Character.isDigit(ch))
         {
            return false;
         }
         
         // special case where we are not allowed to insert digits if there
         // is no more room (we must exclude the special extra digit that
         // is there to allow the re-purposing of the sign char as a digit
         // in some obscure situations)
         if (extra && digits[1] != LEADING_ZERO && (digits[1] != 0 || digits[0] != LEADING_ZERO))
         {
            return false;
         }

         ScreenCharInfo leftInfo = null;
         int leftPos = Integer.MAX_VALUE; // to keep compiler happy
         int leftIndex = -1;
         
         // flag to allow the decimal point to be found on the left side of
         // the cursor before any digit
         boolean leftPointNeigh = false;

         int leftCursorPos = Math.min(presCursorPos, screen.length() - 1);
         while (leftCursorPos >= 0)
         {
            ScreenCharInfo info = screenInfo[leftCursorPos];
            int type = info.getType();
            if (type == ScreenCharInfo.DIGIT)
            {
               leftInfo = info;
               leftPos = leftInfo.getDigitPos();
               leftIndex = calcDigitIndex(leftPos);
               break;
            }
            else if (type == ScreenCharInfo.DEC_SEP)
            {
               leftPointNeigh = true;
            }
            leftCursorPos--;
         }

         int newIndex;
         
         // flag to note if the final cursor position needs adjusting since
         // it should be after the inserted digit but the overwrite digit
         // worker method won't leave it there always
         boolean shiftedCursor = false;
         
         if (leftInfo == null || digits[leftIndex] == LEADING_ZERO)
         {
            newIndex = insertLeadingPlace();
            if (newIndex < 0)
            {
               // no room to insert
               return false;
            }
         }
         else if (leftCursorPos < presCursorPos && 
                  leftPos == -scale             &&
                  (scale > 0 || leftPointNeigh))
         { 
            // trailing insert
            if (scale == fmtFractDigitsCount + fmtFractSharedCount ||
                digits[0] != LEADING_ZERO)
            {
               // no room to insert (digits[0] was checked for LEADING_ZERO
               // not for 0 because we need an unused '>' place)
               return false;
            }

            for (int i = 0; i < digits.length - 1; i++)
            {
               digits[i] = digits[i + 1];
            }
            scale++;

            newIndex = -scale;
            shiftedCursor = true;
         }
         else
         { 
            // insert into a middle after current digit
            if (digits[0] != LEADING_ZERO && digits[0] != 0)
            { 
               // no room before number, try to contract fractional part
               if (scale == fmtFractDigitsCount ||
                   digits[digits.length - 1] != 0)
               {
                  // no room to insert
                  return false;
               }

               newIndex = leftIndex + 1;
               for (int i = digits.length - 1; i > newIndex; i--)
               {
                  digits[i] = digits[i - 1];
               }
               scale--;
            }
            else
            {
               int index = 0;
               while (index <= leftIndex)
               {
                  if (digits[index] == LEADING_ZERO || digits[index] == 0)
                  {
                     index++;
                  }
                  else
                  {
                     break;
                  }
               }

               while (index <= leftIndex)
               {
                  digits[index - 1] = digits[index];
                  index++;
               }

               newIndex = leftIndex;
            }
         }

         // we made room for this digit, write it into the array (note that
         // this sets the cursor too)
         overwriteDigit(newIndex, ch);
         
         // fixup the known cursor position if needed
         if (shiftedCursor)
         {
            presCursorPos++;
         }

         return true;
      }

      /**
       * Inputs character in overwrite mode.
       * 
       * @param    ch
       *           Input character
       * 
       * @return   <code>true</code> if input is accepted (nevertheless it
       *           may be silently ignored in reality).
       */
      private boolean overwrite(char ch)
      {
         // accepting char the same with one in cursor position
         if (!Character.isDigit(ch))
         {
            if (presCursorPos >= screen.length() ||
                screen.charAt(presCursorPos) != ch)
            {
               return false;
            }

            presCursorPos++;
            return true;
         }

         // handling digits
         ScreenCharInfo rightInfo = null;
         int rightPos = Integer.MIN_VALUE; // to keep compiler happy
         int rightIndex = -1;

         int rightCursorPos = presCursorPos;
         while (rightCursorPos < screen.length())
         {
            ScreenCharInfo info = screenInfo[rightCursorPos];
            int type = info.getType();
            if (type == ScreenCharInfo.DIGIT)
            {
               rightInfo = info;
               rightPos = rightInfo.getDigitPos();
               rightIndex = calcDigitIndex(rightPos);
               break;
            }
            rightCursorPos++;
         }

         int newIndex;
         if (rightInfo == null)
         { 
            // trailing insert
            if (fmtHavePeriod)
            { 
               // insert into end of fractional part
               if (scale >= fmtFractDigitsCount + fmtFractSharedCount ||
                   digits[0] != LEADING_ZERO)
               {
                  // no room to append
                  return false;
               }

               for (int i = 0; i < digits.length - 1; i++)
               {
                  digits[i] = digits[i + 1];
               }
               scale++;
               newIndex = digits.length - 1;
            }
            else
            { 
               // number is integer-type
               int index = 0;
               while (index < digits.length &&
                      (digits[index] == LEADING_ZERO || digits[index] == 0))
               {
                  index++;
               }

               if (index == 0)
               { 
                  // no room to append
                  return false;
               }

               while (index < digits.length)
               {
                  digits[index - 1] = digits[index];
                  index++;
               }

               newIndex = digits.length - 1 - scale;
            }
         }
         else
         { 
            // overwrite some existing digit (possibly leading zero)
            if (digits[rightIndex] == LEADING_ZERO ||
                rightCursorPos > presCursorPos && rightPos >= 0 &&
                (rightIndex == 0 || digits[rightIndex - 1] == LEADING_ZERO))
            {
               newIndex = insertLeadingPlace();
               if (newIndex < 0)
               {
                  // no room to insert so overwriting
                  newIndex = rightIndex;
               }
            }
            else
            { 
               // simple overwrite
               newIndex = rightIndex;
            }
         }

         overwriteDigit(newIndex, ch); // this sets cursor also
         presCursorPos++;

         return true;
      }

      /**
       * Calculates index where new leading digit may be placed. Contracts
       * fractional part if necessary.
       * 
       * @return  Non negative index for new digit or -1 if no room available.
       */
      private int insertLeadingPlace()
      {
         int index = 0;
         int endIndex = digits.length - scale;
         
         // insert over last leading zero, except when the first digit is
         // zero (in this case insert over the first digit)
         while (index < endIndex)
         {
            if (digits[index] != LEADING_ZERO)
            {
               break;
            }
            index++;
         }

         if (index == 0)
         {
            if (digits[index] != 0)
            {
               // the integer part is filled, try to make room for a new
               // digit by clearing the last zero in the fractional part
               if (scale == fmtFractDigitsCount ||
                   digits[digits.length - 1] != 0)
               {
                  // no room to insert
                  return -1;
               }

               // shift (copy) the digits to the right by 1 digit
               for (int i = digits.length - 1; i > 0; i--)
               {
                  digits[i] = digits[i - 1];
               }
               
               // move the decimal point to the right
               scale--;
            }
         }
         else if (index == endIndex || digits[index] != 0)
         { 
            // all int digits are leading zeros or first digit is nonzero
            index--;
         }

         return index;
      }

      /**
       * Overwrites digit at given index with given char. Sets cursor to this
       * overwritten place. Sets <code>lastCharEntered</code> flag if this 
       * state is detected.
       * 
       * @param  index
       *         Index in digits[] array.
       * @param  ch
       *         Char which must be decimal digit.
       */
      private void overwriteDigit(int index, char ch)
      {
         // edit the data
         digits[index] = (byte) Character.digit(ch, 10);
         
         // regenerate the display string AND the screen info array
         renderScreen();

         // reset our cursor position
         presCursorPos = 0;
         
         // calculate the digit position
         int newPos = digits.length - 1 - scale - index;
         
         // scan the screen info array to find the cursor position that 
         // matches the digit position that we should edit, this is done so
         // that we do not "miss" a new digit (that may have been inserted
         // since this common code is called from insert mode too)
         while (presCursorPos < screenInfo.length)
         {
            ScreenCharInfo info = screenInfo[presCursorPos];
            
            int type = info.getType();
            int dpos = info.getDigitPos();
            
            if (type == ScreenCharInfo.DIGIT && dpos == newPos)
            {
               break;
            }
            
            presCursorPos++;
         }
         
         // check if a last digit was entered
         if (index == digits.length - 1)
         {
            // This will be real last digit if an one of these conditions
            // apply:
            // 1) No fractional part and no leading zeros. This means that we
            //    can't shift the digits to insert the next one.
            // 2) A last digit in the fractional part has been overwritten.
            //    To examine such condition we must calculate the current
            //    maximum length of the fractional part.

            byte d = digits[0];
            boolean noLeadingZero = (d != LEADING_ZERO && d != 0);

            int fractMaxLen = fmtFractDigitsCount + fmtFractSharedCount;
            if (fractMaxLen == 0)
            {
               if (noLeadingZero)
               {
                  setLastCharEntered(true);
               }
            }
            else
            {
               if (scale < fractMaxLen)
               {
                  // this is possible only when fmtFractSharedCount > 0
                  if (noLeadingZero)
                  {
                     fractMaxLen = scale;
                  }
                  // otherwise just keeping fractMaxLen > scale
                  // we don't need to calculate an exact value
               }

               if (scale == fractMaxLen)
               {
                  setLastCharEntered(true);
               }
            }
         }
      }

      /**
       * Fills screen cache accordingly to current internal values. This
       * includes not only displayed chars but their types (digit, sign etc.)
       * also.
       */
      private void renderScreen()
      {
         if (digits == null)
         {
            screen = "?";

            return;
         }

         if (overflow)
         {
            screen = StringHelper.repeatChar('?', fmtScreenWidth);
            return;
         }
         
         screenBuf.setLength(0);
         
         decimal d = new decimal(0);
         toDecimalVar(d);
         
         char decSep = NumberType.getDecimalSeparator();
         char groupSep = NumberType.getGroupSeparator();

         String fDef = getFormatDef();

         boolean forceSign = (d.compareTo(DECIMALZERO) == 0 && negative);
         
         // let the decimal class create our formatted text (uses a special
         // entry point to tell the decimal class NOT to left pad because
         // the DisplayFormat.toScreenFormat handles this downstream)
         String result = d.toString(fDef, splitFormat, forceSign);

         // remove a leading space inserted for a sign that doesn't exist
         // if the format is between parenthesis (...), then preserver the
         // leading space
         if (isOptionalLeftSign() && 
             !negative && 
             result.charAt(0) == ' ' &&
             fDef.charAt(0) != '(')
         {
            result = result.substring(1);
         }        
         
         int     num    = result.length();
         int     ldigs  = numIntegerChars(result);
         int     pos    = ldigs;
         int     type   = 0;
         boolean left   = true;

         // check if the Z format quirk was applied
         boolean zQuirk = false;
         // index for the digits array
         int     dIdx   = -1;
         // start position for the format definition string
         int     fStart = 0;
         // flag indicating a leading zero was previously forced
         boolean leadingZero = false;
         
         if (fDef.charAt(0) == '+' || fDef.charAt(0) == '-')
         {
            // set the start position for the format definition and the index
            // for the digits array so they are in sync with the rendered 
            // result
            if (num < fDef.length())
            {
               fStart = 1;
            }

            dIdx = 0;
         }
         
         // parse the text and store it as our internal data structure
         for (int i = 0; i < num; i++)
         {
            char ch  = result.charAt(i);
            char fch = fDef.charAt(i + fStart);
            
            if (Character.isDigit(ch))
            {
               type = ScreenCharInfo.DIGIT;
               
               // this will write positioning data too, in descending order
               // from (pos - 1) down to 0 for digits on the left side of
               // the decimal point and -1 downwards for digits on the right
               // side
               pos--;
               addScreenChar(ch, type, pos);
               
               left = false;

               // mark first positive digit encounter - no more leading zeros
               // allowed
               leadingZero = false;
               continue;
            }
            else if (ch == decSep || (leadingZero && fch == decSep))
            {
               type  = ScreenCharInfo.DEC_SEP;
               left  = false;

               // allow decimal separators for leading zeros
               if (leadingZero)
               {
                  ch = fch;
               }
            }
            else if (ch == groupSep || (leadingZero && fch == groupSep))
            {
               type = ScreenCharInfo.GROUP_SEP;

               // allow grouping for leading zeros
               if (leadingZero)
               {
                  ch = fch;
               }
            }
            else
            {
               type = left ? ScreenCharInfo.STR_LEFT
                           : ScreenCharInfo.STR_RIGHT;
            }

            // quirk for the 'z' format - allow leading 0 digits; if after
            // deleting the first digit the number is changed from "30000" to
            // "0000" and the format is "zzzzz", then the "0000" chars are
            // displayed. if the '+' or '-' key was pressed, the leading
            // zeros must be cleared
            if (!clearLeadingZeros && (fch == 'z' || fch == 'Z'))
            {
               dIdx = dIdx + 1;
               if (digits[dIdx] == 0        && 
                   digits[dIdx] != ch - '0' && 
                   ch == ' ')
               {
                  // if the rendered char is blank and the actual digit is
                  // zero, for output of a leading zero
                  ch   = '0';
                  type = ScreenCharInfo.DIGIT;

                  zQuirk      = true;
                  leadingZero = true;
               }
            }
            
            addScreenChar(ch, type);
         }
         
         // no left digits in output and the format string is all suppressed
         // on the left side. if the 'z' quirk was applied, this has nothing
         // else to do - the 0 digit was added by the quirk.
         if (!zQuirk && ldigs == 0 && isZeroSuppressed() && isLeftOnly())
         {
            // check if the rightmost integer digit is a zero instead of a leading zero
            if (isExplicitZero())
            {
               // display a zero where normal decimal class formatting didn't
               // place one (take into account that there may be padded right
               // side sign characters)
               int last  = screenBuf.lastIndexOf(" ");
               int rsign = negative ? 0 : rightSignSize(); 
               last -= rsign;
               
               if (last >= 0)
               {
                  // normal path
                  screenBuf.replace(last, last + 1, "0");
               }
            }
         }
         
         screen = screenBuf.toString();
      }
      
      /**
       * Determines if the rightmost integer digit (just to the left of any
       * decimal point) is an explicit zero.
       *
       * @return   <code>true</code> for an explicit zero, <code>false</code>
       *           for a leading zero or any other digit.
       */
      private boolean isExplicitZero()
      {
         int least = digits.length - 1 - scale;
         return (digits[least] == 0);
      }
      
      /**
       * Calculate the number of digits to the left of the decimal point
       * (this is the same as the total number of digits for any number
       * that has no decimal point or no digits to the right of the decimal
       * point).
       *
       * @param    text
       *           The formatted number as a string.
       *
       * @return   The number of integer digits.
       */
      private int numIntegerChars(String text)
      {
         int num = 0;
         
         char decSep = NumberType.getDecimalSeparator();

         int point = text.indexOf(decSep);
         int end   = point < 0 ? text.length() : point;
         
         for (int i = 0; i < end; i++)
         {
            if (Character.isDigit(text.charAt(i))) 
            {
               num++;
            }
         }
         
         return num;
      }

      /**
       * Appends non digit char to screen cache buffer.
       * 
       * @param   ch
       *          Character.
       * @param   type
       *          Type of this character. Constant from
       *          <code>ScreenCharInfo</code> class.
       */
      private void addScreenChar(char ch, int type)
      {
         int index = screenBuf.length();
         screenBuf.append(ch);
         try
         {
            screenInfo[index].setType(type);
         }
         catch (RuntimeException x)
         {
            throw x;
         }
      }

      /**
       * Appends digit char to screen cache buffer.
       * 
       * @param   ch
       *          Character.
       * @param   type
       *          Type of this character. Always
       *          <code>ScreenCharInfo.DIGIT</code>.
       * @param   digitPos
       *          Position of this digit.
       */
      private void addScreenChar(char ch, int type, int digitPos)
      {
         int index = screenBuf.length();
         screenBuf.append(ch);
         screenInfo[index].setType(type);
         screenInfo[index].setDigitPos(digitPos);
      }

      /**
       * Appends non digit char to screen cache buffer.
       * 
       * @param   str
       *          String.
       * @param   type
       *          Type of this string. It will be type for each character from
       *          it. Constant from <code>ScreenCharInfo</code> class.
       */
      //TODO: method is unused
      @SuppressWarnings("unused")
      private void addScreenString(String str, int type)
      {
         int index = screenBuf.length();
         screenBuf.append(str);
         for (int i = 0; i < str.length(); i++)
         {
            screenInfo[index++].setType(type);
         }
      }

      /**
       * Clears screen cache so it will be refilled on the next reading.
       */
      private void invalidateScreen()
      {
         screen = null;
      }

      /**
       * Refills screen cache if it is invalid.
       */
      private void validateScreen()
      {
         if (screen == null)
         {
            renderScreen();
         }
      }
      
      /**
       * Perform common initialization of the byte array.
       */
      private void initArray()
      {
         // certain signs actually extend the number of valid digits for a
         // non-signed value
         extra = (fs.signLeft == NumberType.SIGN_LEFT_MINUS  ||
                  fs.signLeft == NumberType.SIGN_LEFT_PLUS   ||
                  fs.signRight == NumberType.SIGN_RIGHT_MINUS ||
                  fs.signRight == NumberType.SIGN_RIGHT_PLUS);
         // size of the array
         int arraySize = fmtIntSharedCount   +
                         fmtIntDigitsCount   +
                         fmtFractDigitsCount;
                         
         if (extra)
            arraySize++;

         // allocate the byte array
         digits = new byte[arraySize];         
      }

      /**
       * Additional info about character in screen value.
       */
      @SuppressWarnings("hiding")
      private class ScreenCharInfo
      {
         /** Character is part of left substring. */
         public static final int STR_LEFT = 1;
         
         /** Character is a group separator. */
         public static final int GROUP_SEP = 2;
         
         /** Character is decimal separator character. */
         public static final int DEC_SEP = 3;
         
         /** Character is a digit. */
         public static final int DIGIT = 4;
         
         /** Character is part of right substring. */
         public static final int STR_RIGHT = 6;

         /** Character type. */
         private int type;

         /** First digit position. */
         private int digitPos;

         /**
          * Get type information.
          * 
          * @return  Type field value.
          */
         public int getType()
         {
            return type;
         }

         /**
          * Set character type.
          * 
          * @param   newType
          *          New type of the character.
          */
         public void setType(int newType)
         {
            type = newType;
         }

         /**
          * Return position of the digit character.
          * 
          * @return  Digit position.
          */
         public int getDigitPos()
         {
            return digitPos;
         }

         /**
          * Set new position of digit.
          * 
          * @param   pos
          *          New digit position.
          */
         public void setDigitPos(int pos)
         {
            digitPos = pos;
         }
      }
   }
}
